{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpUtils.pas. The Original Code is available through    }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit contains general purpose conversion routines and testing for       }
{ special values, as well as synchronization utilities.                        }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 July 18                                                 }
{                                                                              }
{ Dependencies  : Windows, SysUtils, Classes (pervasive)                       }
{                                                                              }
{ Version       : 0.3  03Jul18 Added TDSpQueue, FloatsSameInv                  }
{                 0.2  03May09 Added support for mutexes and file mapping.     }
{                 0.1  03Apr17 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpUtils;
interface
uses Windows, SysUtils, Classes;

type
  TIntegerArray = array of integer;
  
  EFileMapping = class(Exception);

  TMutex = class
  private
    FHandle   : THandle;
    FTimeOut  : cardinal;
  public
    constructor Create(const name: string);
    destructor  Destroy; override;

    function    Get: boolean;
    function    Release: boolean;
    property    TimeOut: cardinal read FTimeOut write FTimeOut;
  end;

  TFileMapping = class;

  TFileView = class
  private
    FFileMap  : TFileMapping;
    FMemory   : pointer;
    FCurrent  : pointer;
    FPosition : cardinal;
    FMutex    : TMutex;
    FSize     : cardinal;
  private
    procedure   SetPosition(pos: cardinal);
  public
    constructor Create(map: TFileMapping; mtx: TMutex; access, size, start: cardinal; starthi: cardinal = 0);
    destructor  Destroy; override;

    function    Read(buffer: pointer; count: cardinal): cardinal;
    function    Write(buffer: pointer; count: cardinal): cardinal;
    function    WriteSafe(buffer: pointer; count: cardinal): cardinal;
    function    Flush: boolean;

    function    Lock: boolean;
    procedure   Unlock;

    property    BaseAddress: pointer read FMemory;
    property    Position: cardinal read FPosition write SetPosition;
  end;

  TFileMapping = class
  private
    FFileHandle : THandle;
    FMapHandle  : THandle;
    FSize       : cardinal;
    FHiSize     : cardinal;
    FPageSize   : cardinal;
    FProtect    : cardinal;
    FShared     : boolean;
  public
    constructor Create(const fn, name: string; sec: PSecurityAttributes;
                       protect: cardinal; size: cardinal; hisize: cardinal = 0);
    destructor  Destroy; override;

    function    MakeView(mtx: TMutex; start: cardinal; starthi: cardinal = 0;
                         size: cardinal = 0; access: cardinal = 0): TFileView;
  end;

  // TDSpQueue
  PQueueItem = ^TQueueItem;
  TQueueItem = record
    Item: pointer;
    Next: PQueueItem;
  end;

  TDSpQueue = class
  protected
    FHead, FTail: PQueueItem;
    FLength     : integer;
  public
    constructor Create;
    destructor  Destroy; override;

    procedure   Put(x: pointer);
    function    Get: pointer;
    function    Peek: pointer;
    procedure   HeadToTail;
    function    Empty: boolean;

    property    Length: integer read FLength;
  end;

procedure ValidateIndex(const low, high, ndx: integer);
function  FloatsSame(const a, b: double): boolean;
function  FloatsSameInv(const a, binv: double): boolean;

function IntMax(a, b: integer) : integer;
function IntMin(a, b: integer) : integer;
function CardMax(a, b: cardinal) : cardinal;
function CardMin(a, b: cardinal) : cardinal;

function NaNextended : extended;
function NaNdouble : double;
function NaNsingle : single;

function IsNan(v : extended) : boolean; overload;
function IsNaN(v : double) : boolean; overload;
function IsNaN(v : single) : boolean; overload;

function EndianSwap(n : integer) : integer; overload;
function EndianSwap(n : cardinal) : cardinal; overload;
function EndianSwap(n : word)    : word; overload;

procedure Swap2Bytes(const x; n : integer); register;
procedure Swap4Bytes(const x; n : integer); register;
procedure SwapBytes(var x : cardinal); overload;
procedure SwapBytes(var x : word); overload;
procedure SwapBytes(var x : smallint); overload;

implementation
///////////////////////////////////////////////////////////////////////////////
// TMutex
constructor TMutex.Create(const name: string);
begin
  FHandle := CreateMutex(nil, false, PChar(name));
  if FHandle = 0 then
    raise EFileMapping.CreateFmt('Could not create mutex %s', [name]);

  FTimeOut := 1000; // 1 second default time out
end;

destructor TMutex.Destroy;
begin
  Release;
  CloseHandle(FHandle);
end;

function TMutex.Get: boolean;
begin
  Result := WaitForSingleObject(FHandle, FTimeOut) = WAIT_OBJECT_0;
end;

function TMutex.Release: boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

//////////////////////////////////////////////////////////////////////////////
// TFileMapping
constructor TFileMapping.Create(const fn, name: string; sec: PSecurityAttributes;
                                protect, size: cardinal; hisize: cardinal = 0);
// Create a file mapping object from a file name or on the page file. The file
// will be opened with access rights compatible with the protect parameter.
// Share mode will be set to 0 for exclusive access. Section bits are not
// checked but passed on the Windows call. If the name parameter is given shared
// use of the views to the mapped object is assumed and a mutex is required for
// reading to and writing from any views based on this file mapping object.
var access : cardinal;
    sysInfo : TSystemInfo;
begin
  FFileHandle := INVALID_HANDLE_VALUE;
  GetSystemInfo(sysInfo);
  FPageSize := sysInfo.dwPageSize;
  FProtect := protect;
  FShared := name <> '';

  // If shared access is requested, check if the named object already exists
  if FShared then
    FMapHandle := OpenFileMapping(8 div (FProtect and $E), false, PChar(name));
  if FMapHandle > 0 then begin
    FSize := size;     // !!!!!!!!!!!!! size must be accurate value, not 0
    FHiSize := hisize;
    Exit;
  end;

  // Open the disk file or use the page file
  if fn = '' then begin  // use the page file
    if size = 0 then
      raise EFileMapping.Create('Invalid size for page file file mapping object');
    FSize := size;
    FHiSize := 0;        // page file use is size limited to 2Gb
  end else begin         // use disk file
    if protect and (PAGE_READWRITE or PAGE_WRITECOPY) > 0 then
      access := GENERIC_READ or GENERIC_WRITE
    else if protect and PAGE_READONLY = PAGE_READONLY then
      access := GENERIC_READ
    else raise EFileMapping.Create('Invalid access mode for file mapping object');

    FFileHandle := CreateFile(PChar(fn), access, 0, sec, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if FFileHandle = INVALID_HANDLE_VALUE then
      raise EFileMapping.CreateFmt('Could not open file %s for file mapping.', [fn]);

    if (size > 0) or (hisize > 0) then begin
      FSize := size;
      FHiSize := hisize;
    end else FSize := GetFileSize(FFileHandle, @FHiSize);
  end;

  try
    FMapHandle := CreateFileMapping(FFileHandle, sec, protect, FHiSize, FSize, PChar(name));
    if FMapHandle = 0 then
      raise EFileMapping.Create('Could not create file mapping object');
  except
    if FFileHandle <> INVALID_HANDLE_VALUE then CloseHandle(FFileHandle);
    raise;
  end;
end;

destructor TFileMapping.Destroy;
begin
  if FMapHandle <> 0 then
    CloseHandle(FMapHandle);

  if FFileHandle <> INVALID_HANDLE_VALUE then
    CloseHandle(FFileHandle);

  inherited;
end;

function TFileMapping.MakeView(mtx: TMutex; start: cardinal; starthi: cardinal = 0; size: cardinal = 0; access: cardinal = 0): TFileView;
// Public function to create a view into a mapped file. Mutex must be used when
// the file mapping object is shared.
begin
  if start mod FPageSize > 0 then
    raise EFileMapping.CreateFmt('Invalid start address %8x for mapped file view', [start]);
  if start + starthi + size > FSize + FHiSize then
    raise EFileMapping.Create('Invalid size for mapped file view');
  if FShared and (mtx = nil) then
    raise EFileMapping.Create('Mutex required for view on shared file mapping object');

  if access = 0 then access := 8 div (FProtect and $E); // inherit access rights

  Result := TFileView.Create(self, mtx, access, size, start, starthi);
end;

//////////////////////////////////////////////////////////////////////////////
// TFileView
constructor TFileView.Create(map: TFileMapping; mtx: TMutex; access, size, start: cardinal; starthi: cardinal = 0);
begin
  FFileMap := map;
  FMutex := mtx;

  if size = 0 then begin
    if map.FHiSize > 0 then
      raise EFileMapping.Create('Mapping object too large for complete view.');
    FSize := map.FSize;
  end else
    FSize := size;

  FMemory := MapViewOfFile(map.FMapHandle, access, starthi, start, size);
  if FMemory = nil then
    raise EFileMapping.Create('View of file mapping object could not be created.');

  FCurrent := FMemory;
  FPosition := 0;
end;

destructor TFileView.Destroy;
begin
  if FMemory <> nil then
    UnmapViewOfFile(FMemory);
end;

procedure TFileView.SetPosition(pos: cardinal);
begin
  if pos >= FSize then
    raise EFileMapping.Create('Invalid position on file view object');

  FPosition := pos;
  FCurrent := pointer(cardinal(FMemory) + pos);
end;

function TFileView.Read(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if (FMutex = nil) or FMutex.Get then begin // must use short bool evaluation!!
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(FCurrent^, buffer^, Result);

    Inc(FPosition, Result);
    Inc(cardinal(FCurrent), Result);

    if FMutex <> nil then FMutex.Release;
  end;
end;

function TFileView.Write(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if (FMutex = nil) or FMutex.Get then begin // must use short bool evaluation!!
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(buffer^, FCurrent^, Result);

    Inc(FPosition, Result);
    Inc(cardinal(FCurrent), Result);

    if FMutex <> nil then FMutex.Release;
  end;
end;

function TFileView.WriteSafe(buffer: pointer; count: cardinal): cardinal;
begin
  Result := 0;
  if (FMutex = nil) or FMutex.Get then begin // must use short bool evaluation!!
    if FPosition + count > FSize then Result := FSize - FPosition
    else Result := count;

    Move(buffer^, FCurrent^, Result);

    try
      if not FlushViewOfFile(FCurrent, Result) then
        raise EFileMapping.CreateFmt('Unable to flush file view at position %d (file address %8x)', [FPosition, FCurrent]);
    finally
      Inc(FPosition, Result);
      Inc(cardinal(FCurrent), Result);

      if FMutex <> nil then FMutex.Release;
    end;
  end;
end;

function TFileView.Flush: boolean;
begin
  if FMutex = nil then Result := true else Result := FMutex.Get;
  if Result then try
    Result := FlushViewOfFile(FMemory, 0);
    if not Result then
      raise EFileMapping.Create('Unable to flush file view');
  finally
    if FMutex <> nil then FMutex.Release;
  end;
end;

function TFileView.Lock: boolean;
begin
  Result := (FMutex = nil) or FMutex.Get;
end;

procedure TFileView.Unlock;
begin
  if FMutex <> nil then FMutex.Release;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpQueue
constructor TDSpQueue.Create;
begin
  inherited;
  FHead := nil;
end;

destructor TDSpQueue.Destroy;
var i : PQueueItem;
begin
  while FHead <> nil do begin   // the items themselves are not deleted
    i := FHead;
    FHead := FHead^.Next;
    Dispose(i);
  end;
end;

procedure TDSpQueue.Put(x: pointer);
var i : PQueueItem;
begin
  New(i);
  i^.Next := nil;
  i^.Item := x;
  if FTail = nil then FHead := i
  else FTail^.Next := i;
  FTail := i;
  Inc(FLength);
end;

function TDSpQueue.Get: pointer;
var i : PQueueItem;
begin
  if FHead = nil then Result := nil
  else begin
    Result := FHead^.Item;
    i := FHead;
    FHead := FHead^.Next;
    if FTail = i then FTail := nil;
    Dispose(i);
    Dec(FLength);
  end;
end;

function TDSpQueue.Peek: pointer;
begin
  if FHead = nil then Result := nil
  else Result := FHead^.Item;
end;

procedure TDSpQueue.HeadToTail;
var i: PQueueItem;
begin
  if (FHead = nil) or (FHead = FTail) then Exit;
  i := FHead;
  FHead := FHead^.Next;
  i^.Next := nil;
  FTail^.Next := i;
  FTail := i;
end;

function TDSpQueue.Empty : boolean;
begin
  Result := FHead = nil;
end;

////////////////////////////////////////////////////////////////////////////////
// utility routines
procedure ValidateIndex(const low, high, ndx: integer);
begin
  if not ndx in [low..high] then
    raise Exception.Create('Invalid index');
end;

function FloatsSame(const a, b: double): boolean;
begin
  if b = 0.0 then Result := a = 0.0
  else Result := Abs(a/b - 1) <= 0.0001;
end;

function FloatsSameInv(const a, binv: double): boolean;
var x: double;
begin
  x := a*binv;
  Result := Abs(x - 1) <= 0.0001;        // binv can be assumed <> 0
end;

function EndianSwap(n : integer) : integer;
// Swaps 32bit integer value in big endian to little endian and vice versa
asm
  BSWAP EAX
end;

function EndianSwap(n : cardinal) : cardinal;
// Swaps 32bit cardinal value in big endian to little endian and vice versa
asm
  BSWAP EAX
end;

function EndianSwap(n : word) : word;
// Swaps 16bit integer value in big endian to little endian and vice versa
asm
  XCHG  AL, AH
end;

// x contains the address to n things of 2 bytes long
procedure Swap2Bytes(const x; n : integer); register;
asm
@NotDone:
  MOV   CX,[EAX]
  XCHG  CH,CL
  MOV   [EAX],CX
  ADD   EAX,2
  DEC   EDX
  JNZ   @NotDone
end;

// x contains the address to n things of 4 bytes long
procedure Swap4Bytes(const x; n : integer); register;
asm
@NotDone:
  MOV   ECX,[EAX]
  BSWAP ECX
  MOV   [EAX],ECX
  ADD   EAX,4
  DEC   EDX
  JNZ   @NotDone
end;

procedure SwapBytes(var x : cardinal);
asm
  BSWAP EAX
end;

procedure SwapBytes(var x : word);
asm
  XCHG  AL, AH
end;

procedure SwapBytes(var x : smallint);
asm
  XCHG  AL, AH
end;

function IntMax(a, b: integer) : integer;
begin
  if a > b then IntMax := a else IntMax := b;
end;

function IntMin(a, b: integer) : integer;
begin
  if a < b then IntMin := a else IntMin := b;
end;

function CardMax(a, b: cardinal) : cardinal;
begin
  if a > b then CardMax := a else CardMax := b;
end;

function CardMin(a, b: cardinal) : cardinal;
begin
  if a < b then CardMin := a else CardMin := b;
end;

function NaNextended : extended;
const b : array [1..10] of byte = ($FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF);
begin
  Move(b, Result, 10);
end;

function NaNdouble : double;
const b : array [1..8] of byte = ($7F, $FF, $FF, $FF, $FF, $FF, $FF, $FF);
begin
  Move(b, Result, 8);
end;

function NaNsingle : single;
const b : array [1..4] of byte = ($7F, $FF, $FF, $FF);
begin
  Move(b, Result, 4);
end;

function IsNan(v : extended) : boolean;
var b : array [1..10] of byte;
    i : integer;
begin
  Move(v, b, 10);
  Result := ((b[1] shl 1) = $FE);
  i := 2;
  while Result and (i <= 10) do begin
    Result := (b[i] = $FF);
    Inc(i);
  end;
end;

function IsNaN(v : double) : boolean;
var i : comp;
begin
  Move(v, i, 8);
  Result := ((i = $7FFFFFFFFFFFFFFF) or (i = $FFFFFFFFFFFFFFFF));
    // No bit-wise operations on comp type possible
end;

function IsNaN(v : single) : boolean;
var i : cardinal;
begin
  Move(v, i, 4);
  Result := ((i shl 1) = $FFFFFFFE); // Discard sign
end;

end.
