unit TreeTransformationList;
(*)
 |
 | Filename: TransformationList.pas
 | Description: Implementation of TTransformationlist, a class that can store
 |              a list of transformation matrices. The collection of matrices
 |              stored in the class can be applied to a vector in order to
 |              transform it.
 |
 |              For example: E * D * C * B * A * v, with A..E matrices and v
 |                           a vector are represented in the class as:
 |
 |                           A = FList[0]
 |                           B = FList[1]
 |                           ...
 |                           E = FList[4]
 |
 | Author: Ton van den Heuvel
 | Date: 15-08-2002
 |
(*)

interface

uses Classes, TreeMatrix, TreeVector;

     {
        TTransformationlist beschrijft een reeks van achtereenvolgende transfor-
        maties die met behulp van de methode TransformVector op een vector
        kunnen worden losgelaten.
     }

type TTransformationList = class(TObject)
     private
       FList: TList;
     public
       constructor Create;
       destructor Destroy; override;
       procedure AddMatrix(const AMatrix: TMatrix);
       function Transform(APoint3d: TPoint3d) : TPoint3d; overload;
       function Transform(AVector3d: TVector3d) : TVector3d; overload;

       procedure ClearList;

       function Count: Integer;
     end;

     {
        Representatieinvariant: TransformationList.t = #t = t.Flist.Count
        Abstractiefunctie: [[t]] = producten van FList in de volgorde
                                   FList.Items[n-1]*FList.Items[n-2]*...*FList.Items[0]

        De abstractiefunctie poogt duidelijk te maken dat de matrices in
        t.FList in feite in omgekeerde volgorde staan ten opzichte van de
        volgorde van vermenigvuldigen tijdens het uitrekenen van een transfor-
        matie.
     }

implementation

constructor TTransformationList.Create;
{ Standaard constructor, die ook nog de transformatie lijst zelf, FList,
  creert. }
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TTransformationList.Destroy;
{ Deze destructor dealloceert al het geheugen ingenomen door de matrices in de
  transformatie lijst en het object zelf. }
var i: Integer;
begin
  for i := 0 to FList.Count - 1 do TMatrix(FList.Items[i]).Free;
  FList.Free;
  inherited Destroy;
end;

procedure TTransformationList.AddMatrix(const AMatrix: TMatrix);
{ AddMatrix voegt de matrix AMatrix toe aan de transformatie lijst. }
var AMatrixCopy: TMatrix;
begin
  AMatrixCopy := TMatrix.Create;
  AMatrixCopy.Assign(AMatrix);
  FList.Add(AMatrixCopy);
end;

function TTransformationList.Transform(APoint3d: TPoint3d) : TPoint3d;
{ Transform is de procedure waar alles om draait binnen het TTransforma-
  tionlist object. In feite worden alle transformatie matrices zoals die op
  het moment van aanroep in de lijst zitten toegepast op AVector. }
var i: Integer;
    DummyPoint3d: TPoint3d;
begin
  Result := TPoint3d.Create;
  Result.Assign(APoint3d);

  for i := 0 to FList.Count - 1 do
  begin
    DummyPoint3d := TMatrix(FList.Items[i]).Multiply(Result);
    Result.Assign(DummyPoint3d);
    DummyPoint3d.Destroy;
  end;
end;

function TTransformationList.Transform(AVector3d: TVector3d) : TVector3d;
{ Transform is de procedure waar alles om draait binnen het TTransforma-
  tionlist object. In feite worden alle transformatie matrices zoals die op
  het moment van aanroep in de lijst zitten toegepast op AVector. }
var i: Integer;
    DummyVector3d: TVector3d;
begin
  Result := TVector3d.Create;
  Result.Assign(AVector3d);

  for i := 0 to FList.Count - 1 do
  begin
    DummyVector3d := TMatrix(FList.Items[i]).Multiply(Result);
    Result.Assign(DummyVector3d);
    DummyVector3d.Destroy;
  end;
end;

procedure TTransformationList.ClearList;
{ ClearList verwijdert alle matrices uit de transformatie lijst, en dealloceert
  bovendien ook het geheugen dat in beslag wordt genomen door deze matrices. }
var i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    TMatrix(FList.Items[i]).Free;
  FList.Clear;
end;

function TTransformationList.Count: Integer;
{ Count retourneert het aantal matrices in de transformatie lijst. }
begin
  Count := FList.Count;
end;

end.
