unit TreeVector;
(*
 | Deze unit bevat definities van klassen van zowel 3d punten als
 | 3d vectoren. Bovendien zijn er functies opgenomen die standaard operaties op
 | vectoren implementeren, zoals bijvoorbeeld het inproduct nemen van twee
 | vectoren.
 |                                                       (c) Ton van den Heuvel
 |                                                           13-04-2001
 *)
interface

uses Classes;

type TVector3d = class;

     TPoint3d = class
     protected
       { Note: we use homogeneous coordinates }
       FData: array[0..3] of real;
       function GetData(i: byte): real;
       { Pre: 0 <= i < 3
         Ret: Data[i] }
       procedure SetData(i: byte; Value: real);
       { Pre: 0 <= i < 3
         Post: FData[i] = Value }
     public
       property x: real read FData[0] write FData[0];
       property y: real read FData[1] write FData[1];
       property z: real read FData[2] write FData[2];
       property Data[i: byte]: real read GetData write SetData; default;

       constructor Create; overload;
       constructor Create(Ax, Ay, Az: real); overload;

       destructor Destroy; override;

       function Add(const C: real): TPoint3d; overload;
       { Pre: True
         Ret: Result = Self + C }
       function Add(const APoint3d: TPoint3d): TPoint3d; overload;
       { Pre: AVector3d <> nil
         Ret: Result.x = Self.x + APoint3d.x, Result.y = Self.y + APoint3d.y and
              Result.z = Self.z + APoint3d.z }
       function Add(const AVector3d: TVector3d): TPoint3d; overload;
       { Pre: AVector3d <> nil
         Ret: Result.x = Self.x + AVector3d.x, Result.y = Self.y + AVector3d.y and
              Result.z = Self.z + AVector3d.z }

       function Translate(const AVector3d: TVector3d; const Factor: real): TPoint3d;
       { Pre: AVector3d <> nil
         Post: Result.x = Self.x + AVector3d.x * Factor, Result.y = Self.y + AVector3d.y * Factor and
               Result.z = Self.z + AVector3d.z * Factor }

       procedure Assign(const ASourcePoint3d: TPoint3d); overload;
       procedure Assign(const ASourceVector3d: TVector3d); overload;
     end;

     TVector3d = class
     private
       { Note: We use homogeneous coordinates }
       FData: array[0..3] of real;
       function GetData(i: byte): real;
       { Pre: 0 <= i <= 3
         Ret: FData[i] }
       procedure SetData(i: byte; Value: real);
       { Pre: 0 <= i <= 3
         Post: FData[i] = Value }
     public
       property x: real read FData[0] write FData[0];
       property y: real read FData[1] write FData[1];
       property z: real read FData[2] write FData[2];
       property Data[i: byte]: real read GetData write SetData; default;

       constructor Create; overload;
       constructor Create(const Ax, Ay, Az: Real); overload;
       constructor Create(const AVector1, AVector2: TVector3d); overload;
       constructor Create(const APoint1, APoint2: TPoint3d); overload;

       function Add(const C: Real): TVector3d; overload;
       function Add(const APoint: TPoint3d): TVector3d; overload;
       function Add(const AVector: TVector3d): TVector3d; overload;
       function Subtract(const C: Real): TVector3d; overload;
       function Subtract(const APoint: TPoint3d): TVector3d; overload;
       function Subtract(const AVector: TVector3d): TVector3d; overload;
       function Multiply(const C: Real): TVector3d; overload;
       function Multiply(const APoint: TPoint3d): TVector3d; overload;
       function Multiply(const AVector: TVector3d): TVector3d; overload;
       function Divide(const C: real): TVector3d; overload;
       function Divide(const APoint: TPoint3d): TVector3d; overload;
       function Divide(const AVector: TVector3d): TVector3d; overload;
       function CrossProduct(const AVector: TVector3d): TVector3d;
       function Normal: TVector3d;
       function Length: Real;
       function Equal(const AVector: TVector3d): Boolean;
       function Angle(const AVector: TVector3d): Real;
       function DotProduct(const AVector: TVector3d): Real;

       procedure Normalize;
       procedure Abs;
       procedure ZeroVector;
       procedure Assign(const ASourceVector: TVector3d); overload;
       procedure Assign(const AVector1, AVector2: TVector3d); overload;
     end;

var UnitX, UnitY, UnitZ: TVector3d;         // Standard x, y and z unit vectors
    StandardOrigin: TPoint3d;

implementation

uses Math;

{ TPoint3d }

constructor TPoint3d.Create;
begin
  inherited Create;
  FData[0] := 0;
  FData[1] := 0;
  FData[2] := 0;
  FData[3] := 1;
end;

constructor TPoint3d.Create(Ax, Ay, Az: real);
begin
  inherited Create;
  FData[0] := Ax;
  FData[1] := Ay;
  FData[2] := Az;
  FData[3] := 1;
end;

function TPoint3d.GetData(i: byte): real;
{ Pre: 0 <= i < 3
  Ret: FData[i] }
begin
  Result := FData[i];
end;

procedure TPoint3d.SetData(i: byte; Value: real);
{ Pre: 0 <= i < 3
  Post: FData[i] = Value }
begin
  FData[i] := Value;
end;

destructor TPoint3d.Destroy;
begin
  inherited Destroy;
end;

function TPoint3d.Add(const C: real): TPoint3d;
{ Pre: True
  Ret: Result = Self + C }
begin
  Result := TPoint3d.Create;
  Result.x := Self.x + C;
  Result.y := Self.y + C;
  Result.z := Self.z + C;
end;

function TPoint3d.Add(const APoint3d: TPoint3d): TPoint3d;
{ Pre: AVector3d <> nil
  Ret: Result.x = Self.x + APoint3d.x, Result.y = Self.y + APoint3d.y and
       Result.z = Self.z + APoint3d.z }
begin
  Result := TPoint3d.Create;
  Result.x := Self.x + APoint3d.x;
  Result.y := Self.y + APoint3d.y;
  Result.z := Self.z + APoint3d.z;
end;

function TPoint3d.Add(const AVector3d: TVector3d): TPoint3d;
{ Pre: AVector3d <> nil
  Ret: Result.x = Self.x + AVector3d.x, Result.y = Self.y + AVector3d.y and
       Result.z = Self.z + AVector3d.z }
begin
  Result := TPoint3d.Create;
  Result.x := Self.x + AVector3d.x;
  Result.y := Self.y + AVector3d.y;
  Result.z := Self.z + AVector3d.z;
end;

function TPoint3d.Translate(const AVector3d: TVector3d; const Factor: real): TPoint3d;
{ Pre: AVector3d <> nil
  Post: Result.x = Self.x + AVector3d.x * Factor, Result.y = Self.y + AVector3d.y * Factor and
        Result.z = Self.z + AVector3d.z * Factor }
begin
  Result := TPoint3d.Create;
  Result.x := Self.x + AVector3d.x * Factor;
  Result.y := Self.y + AVector3d.y * Factor;
  Result.z := Self.z + AVector3d.z * Factor;
end;

procedure TPoint3d.Assign(const ASourcePoint3d: TPoint3d);
begin
  FData[0] := ASourcePoint3d.x;
  FData[1] := ASourcePoint3d.y;
  FData[2] := ASourcePoint3d.z;
end;

procedure TPoint3d.Assign(const ASourceVector3d: TVector3d);
begin
  FData[0] := ASourceVector3d.x;
  FData[1] := ASourceVector3d.y;
  FData[2] := ASourceVector3d.z;
end;

{ TVector3d }

constructor TVector3d.Create;
{ Standaard constructor die een nul-vector creert. }
begin
  inherited Create;
  FData[0] := 0;
  FData[1] := 0;
  FData[2] := 0;
  FData[3] := 0;
end;

constructor TVector3d.Create(const Ax, Ay, Az: Real);
{ Deze constructor creert een vector vanuit de oorsprong naar de opgegeven
  cordinaten in de drie dimensionale ruimte. }
begin
  inherited Create;
  FData[0] := Ax;
  FData[1] := Ay;
  FData[2] := Az;
  FData[3] := 0;
end;

constructor TVector3d.Create(const AVector1, AVector2: TVector3d);
{ Deze constructor creert een vector van AVector2 naar AVector1. }
begin
  inherited Create;
  FData[0] := AVector1.x - AVector2.x;
  FData[1] := AVector1.y - AVector2.y;
  FData[2] := AVector1.z - AVector2.z;
  FData[3] := 0;
end;

constructor TVector3d.Create(const APoint1, APoint2: TPoint3d);
{ Deze constructor crert een vector van APoint2 naar APoint1. }
begin
  inherited Create;
  FData[0] := APoint1.x - APoint2.x;
  FData[1] := APoint1.y - APoint2.y;
  FData[2] := APoint1.z - APoint2.z;
  FData[3] := 0;
end;

function TVector3d.GetData(i: byte): real;
{ Pre: 0 <= i < 3
  Ret: FData[i] }
begin
  Result := FData[i];
end;

procedure TVector3d.SetData(i: byte; Value: real);
{ Pre: 0 <= i < 3
  Post: FData[i] = Value }
begin
  FData[i] := Value;
end;

function TVector3d.Add(const C: Real): TVector3d;
{ Add retourneert de vector Self met daarbij opgeteld de constante C. Het re-
  sultaat wordt via Result geretourneerd. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] + C;
  Result.y := Self.FData[1] + C;
  Result.z := Self.FData[2] + C;
end;

function TVector3d.Add(const APoint: TPoint3d): TVector3d;
{ Add retourneert de vector Self met daarbij opgeteld het punt APoint. Het re-
  sultaat wordt via Result geretourneerd. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] + APoint.x;
  Result.y := Self.FData[1] + APoint.y;
  Result.z := Self.FData[2] + APoint.z;
end;

function TVector3d.Add(const AVector: TVector3d): TVector3d;
{ Add retourneert de vector Self met daarbij de vector AVector opgeteld. Het
  resultaat wordt via Result geretourneerd. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] + AVector.x;
  Result.y := Self.FData[1] + AVector.y;
  Result.z := Self.FData[2] + AVector.z;
end;

function TVector3d.Subtract(const C: Real): TVector3d;
{ Subtract retourneert via Result het verschil tussen de vector Self en de
  constante C. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] - C;
  Result.y := Self.FData[1] - C;
  Result.z := Self.FData[2] - C;
end;

function TVector3d.Subtract(const APoint: TPoint3d): TVector3d;
{ Subtract retourneert via Result het verschil tussen de vector Self en de
  cordinaten van het punt APoint. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] - APoint.x;
  Result.y := Self.FData[1] - APoint.y;
  Result.z := Self.FData[2] - APoint.z;
end;

function TVector3d.Subtract(const AVector: TVector3d): TVector3d;
{ Subtract retourneert via Result het verschil tussen de vectoren Self en
  AVector. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] - AVector.x;
  Result.y := Self.FData[1] - AVector.y;
  Result.z := Self.FData[2] - AVector.z;
end;

function TVector3d.Multiply(const C: Real): TVector3d;
{ Multiply retourneert via Result het product van de vector Self en de constante
  C. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] * C;
  Result.y := Self.FData[1] * C;
  Result.z := Self.FData[2] * C;
end;

function TVector3d.Multiply(const APoint: TPoint3d): TVector3d;
{ Multiply retourneert via Result het product van de vector Self en de
  cordinaten van het punt APoint. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] * APoint.x;
  Result.y := Self.FData[1] * APoint.y;
  Result.z := Self.FData[2] * APoint.z;
end;

function TVector3d.Multiply(const AVector: TVector3d): TVector3d;
{ Multiply retourneert via Result het product van de twee vectoren Self en
  AVector. }
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] * AVector.x;
  Result.y := Self.FData[1] * AVector.y;
  Result.z := Self.FData[2] * AVector.z;
end;

function TVector3d.Divide(const C: real): TVector3d;
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] / C;
  Result.y := Self.FData[1] / C;
  Result.z := Self.FData[2] / C;
end;

function TVector3d.Divide(const APoint: TPoint3d): TVector3d;
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] / APoint.x;
  Result.y := Self.FData[1] / APoint.y;
  Result.z := Self.FData[2] / APoint.z;
end;

function TVector3d.Divide(const AVector: TVector3d): TVector3d;
begin
  Result := TVector3d.Create;
  Result.x := Self.FData[0] / AVector.x;
  Result.y := Self.FData[1] / AVector.y;
  Result.z := Self.FData[2] / AVector.z;
end;

procedure TVector3d.Normalize;
{ Normalize maakt van de huidige vector een eenheidsvector. }
var l: Real;
begin
  l := Length;
  if l <> 0 then
  begin
    FData[0] := FData[0] / l;
    FData[1] := FData[1] / l;
    FData[2] := FData[2] / l;
  end;
end;

procedure TVector3d.Abs;
{ Abs maakt van de vector een absolute vector, i.e. van x, y en z wordt
  de absolute waarde gebruikt. }
begin
  FData[0] := System.Abs(FData[0]);
  FData[1] := System.Abs(FData[1]);
  FData[2] := System.Abs(FData[2]);
end;

procedure TVector3d.ZeroVector;
{ ZeroVector maakt, heel verassend, van de huidige vector een nulvector. }
begin
  FData[0] := 0;
  FData[1] := 0;
  FData[2] := 0;
end;

function TVector3d.CrossProduct(const AVector: TVector3d): TVector3d;
{ CrossProduct rekent het kruisproduct uit van de vectoren Self en AVector, en
  retourneert vervolgens het resultaat via Result. }
begin
  Result := TVector3d.Create;
  Result.x := (FData[1] * AVector.z) - (FData[2] * AVector.y);
  Result.y := (FData[2] * AVector.x) - (FData[0] * AVector.z);
  Result.z := (FData[0] * AVector.y) - (FData[1] * AVector.x);
end;

function TVector3d.DotProduct(const AVector: TVector3d): Real;
{ DotProduct berekent het inproduct van de vectoren Self en AVector en re-
  tourneert vervolgens het resultaat. }
begin
  DotProduct := FData[0] * AVector.x + FData[1] * AVector.y + FData[2] * AVector.z;
end;

procedure TVector3d.Assign(const ASourceVector: TVector3d);
{ Copy wordt gebruikt om exacte kopin van een vector te kunnen maken. }
begin
  FData[0] := ASourceVector.x;
  FData[1] := ASourceVector.y;
  FData[2] := ASourceVector.z;
end;

procedure TVector3d.Assign(const AVector1, AVector2: TVector3d);
begin
  FData[0] := AVector1.x - AVector2.x;
  FData[1] := AVector1.y - AVector2.y;
  FData[2] := AVector1.z - AVector2.z;
end;

function TVector3d.Length: Real;
{ Length retourneert de lengte van de vector Self. }
var l: real;
begin
  l := Sqrt(FData[0] * FData[0] + FData[1] * FData[1] + FData[2] * FData[2]);
  Length := l;
end;

function TVector3d.Normal: TVector3d;
{ Normal retourneert een pointer naar een vector welke de normaal is van de
  vector Self. }
var v: TVector3d;
begin
  v := TVector3d.Create(FData[1], -FData[0], FData[2]);
  Normal := v;
end;

function TVector3d.Equal(const AVector: TVector3d): Boolean;
{ Equal bekijkt of de twee vectoren Self en AVector dezelfde zijn door ze bei-
  den te normaliseren. }
var x1, x2, y1, y2, z1, z2, l1, l2: Real ;
begin
  x1 := 0; x2 := 0; y1 := 0; y2 := 0; z1 := 0; z2 := 0;

  l1 := Self.Length;
  if l1 <> 0 then
  begin
    x1 := FData[0] / l1;
    y1 := FData[1] / l1;
    z1 := FData[2] / l1;
  end;

  l2 := AVector.Length;
  if l2 <> 0 then
  begin
    x2 := AVector.x / l2;
    y2 := AVector.y / l2;
    z2 := AVector.z / l2;
  end;

  if ((l1 <> 0) and (l2 <> 0)) then Equal := ((x1 = x2) and (y1 = y2) and (z1 = z2))
  else Equal := False;
end;

function TVector3d.Angle(const AVector: TVector3d): Real;
{ Angle berekent de hoek tussen de vectoren Self en AVector. }
var l1, l2: Real;
begin
  l1 := Self.Length;
  l2 := AVector.Length;

  if l1 * l2 <> 0 then Angle := ArcCos(Self.DotProduct(AVector) / (l1 * l2))
  else Angle := -1;
end;

initialization

UnitX := TVector3d.Create(1, 0, 0);
UnitY := TVector3d.Create(0, 1, 0);
UnitZ := TVector3d.Create(0, 0, 1);
StandardOrigin := TPoint3d.Create(0, 0, 0);

finalization

UnitX.Destroy;
UnitY.Destroy;
UnitZ.Destroy;
StandardOrigin.Destroy;

end.
