{ ============================================================================= }
{                             UNIT uSun                                         }
{                                                                               }
{ Description: Procedures to models the position and radiance of the sun        }
{ depending the observers position on the earth.                                }
{                                                                               }
{ Created Skinhat June 2004                                                     }
{                                                                               }
{ ============================================================================= }

unit dtmHdsSunUnit;

interface
uses VectorGeometry, sysUtils;

function sunradiance(upAngle:single):single;
procedure sunposition(datetime:TDateTime; latitude,longitude:single; var azsun,altsun:single);
function sunpositionAsVector(distance,azsun,altsun:single):TVector;

implementation

{ ============================================================================= }
{ FUNCTION SunRadiance                                                          }
{ Purpose: Given the angle of the sun from the horizon return the suns radiance.}
{ Will be between 0 and 1                                                       }
{ ============================================================================= }
function sunradiance(upAngle:single):single;
begin
  result:=(1+2*sin(degtorad(upAngle)))/3;
  if result<0 then
    result:=0;
end;


{ ============================================================================= }
{ PROCEDURE SunPosition                                                         }
{ Purpose : Given the observers longitute and latitude, date and time of day    }
{ return the position of the sun. altsun is the angle from the horizon while    }
{ azsun in the angle from north. Modified from QBASIC code with the following   }
{ remarks:                                                                      }
{                                                                               }
{ This program will calculate the position of the Sun                           }
{ using a low precision method found on page C24 of the                         }
{ 1996 Astronomical Almanac.                                                    }
{                                                                               }
{ The method is good to 0.01 degrees in the sky over the                        }
{ period 1950 to 2050.                                                          }
{                                                                               }
{ QBASIC program by Keith Burnett (kburnett@geocity.com)                        }
{                                                                               }
{                                                                               }
{ Work in double precision and define some constants                            }
{                                                                               }
{ ============================================================================= }
procedure sunposition(datetime:TDateTime; latitude,longitude:single; var azsun,altsun:single);
const
  rads = pi / 180;
  tpi = 2 * pi;
  degs = 180 / pi ;

var
  iMonth,iYear,iDay:word;
//  glong,
  glat:single;
  iHour, iMin, iSec, iMSec:word;
  d,lambda:single;
  L,obliq,alpha,delta:single;
//  r,equation:single;
  g:single;
  lmst,hasun:single;
  sinalt,y,x:single;

  //   Get the days to J2000
  //   h is UT in decimal hours
  //   FNday only works between 1901 to 2099 - see Meeus chapter 7
  //
  //DEF FNday (y, m, d, h) = 367 * y - 7 * (y + (m + 9) \ 12) \ 4 + 275 * m\ 9 + d - 730531.5 + h / 24
  //'
  //'   define some arc cos and arc sin functions and a modified inverse
  //'   tangent function
  //'
  function rev( x : real ) : real;
  var rv : real;
  begin
    rv := x - trunc(x/360.0)*360.0;
    if rv < 0.0 then
      rv := rv + 360.0;
    rev := rv;
  end;

  function FNacos(x:single):single;
  var
    s:single;
  begin
    s:= sqrt(1 - x * x);
    result:= radtodeg(arctan(s / x));
  end;

  function FNasin(x:single):single;
  var
    c:single;
  begin
    c:= SQRT(1 - x * x);
    result:= radtodeg(arctan(x / c));
  end;
  //
  //   the atn2 function below returns an angle in the range 0 to two pi
  //   depending on the signs of x and y.
  //
  function FNatn2(y, x:single):single;
  var
    a:single;
  begin
    a:= arctan(y / x);
    IF x < 0 THEN
      a:= a + pi ;
    IF (y < 0) AND (x > 0) THEN
      a:= a + 2*pi;
    result:=radtodeg(a);
  end;

  //
  //   the function below returns the true integer part,
  //   even for negative numbers
  //
  function FNipart(x:single):single;
  begin
    result:= SIGN(x) * INT(ABS(x));
  end;
  //
  //   the function below returns an angle in the range
  //   0 to two pi
  //
  function FNrange(x:single):single;
  var
    b,a:single;
  begin
    b:= x / tpi;
    a:= tpi * (b - FNipart(b));
    IF a < 0 THEN
      a:= tpi + a ;
    result:=a;
  end;
  //
  //   Find the ecliptic longitude of the Sun
  //
  function FNsun(d:single):single;
  begin
  //
  //   mean longitude of the Sun
  //
    L:= radtodeg(FNrange(280.461 * rads + 0.9856474 * rads * d));
  //
  //   mean anomaly of the Sun
  //
    g:= radtodeg(FNrange(357.528 * rads + 0.9856003 * rads * d));
  //
  //   Ecliptic longitude of the Sun
  //
    //result:={radtodeg(FNrange(}L + 1.915 * {rads *} SIN(degtorad(g)) + 0.02 * {rads *} SIN(degtorad(g)*2 ));
    result:=rev(L + 1.915 *  SIN(degtorad(g)) + 0.02 *  SIN(degtorad(g)*2 ));
  //  result:=radtodeg(FNrange(L + 1.915 * rads * SIN(degtorad(g)) + 0.02 * rads * SIN(degtorad(g)*2 )));
  //
  //   Ecliptic latitude is assumed to be zero by definition
  //
  end;

begin
  glat:= degtorad(latitude);  //Used in BOTH
 // glong:= degtorad(longitude);//COS(glat) SIN(glat)

  { We call our day number d. It can be computed from a JD (Julian Day Number) or
   MJD (Modified Julian Day Number) like this:


    d  =  JD - 2451543.5  =  MJD - 51543.0

   We can also compute d directly from the calendar date like this:   }


  DecodeDate(datetime, iYear, iMonth, iDay);

  //DEF FNday (y, m, d, h) = 367 * y - 7 * (y + (m + 9) \ 12) \ 4 + 275 * m\ 9 + d - 730531.5 + h / 24
  DecodeTime(dateTime,iHour,iMin,iSec,iMSec);
  d:= 367*iYear - round(7 * ( iYear + round((iMonth+9)/12) ) / 4) + round(275*iMonth/9) + iDay - 730531.5 ;
 // ihour:=strtoint(inputbox('day','','11'));

  d:=  d+(iHour+iMin/60)/24;

  //
  //   Use FNsun to find the ecliptic longitude of the
  //   Sun
  //
  lambda:= FNsun(d);
  //
  //   Obliquity of the ecliptic
  //
   obliq:= 23.439 {* rads }- 0.0000004 * {rads *} d;
  //
  //   Find the RA and DEC of the Sun
  //
  alpha:= FNatn2(COS(degtorad(obliq)) * SIN(degtorad(lambda)), COS(degtorad(lambda))) ;
  delta:= FNasin(SIN(degtorad(obliq)) * SIN(degtorad(lambda)));
  //
  //   Find the Earth - Sun distance
  //
//  r:= 1.00014 - 0.01671 * COS(degtorad(g)) - 0.00014 * COS(degtorad(g)*2 );
  //
  //   Find the Equation of Time
  //
//   equation:= (L - alpha) * degs * 4;
  //
  //   find the Alt and Az of the Sun for a given position
  //   on Earth
  //
  //   hour angle of Sun
  lmst:=rev(280.46061837 + (360.98564736629 * d) + longitude);
  hasun:= {radtodeg(FNrange(}rev(LMST - alpha);//));
  //
  //   conversion from hour angle and dec to Alt Az
  sinalt:= SIN(degtorad(delta)) * SIN(glat) + COS(degtorad(delta))
                                      //glong ?
                                * COS(glat) * COS(degtorad(hasun)) ;

  altsun:= FNasin(sinalt);

  y:= -COS(degtorad(delta)) * COS(glat) * SIN(degtorad(hasun));
  x:= SIN(degtorad(delta)) - SIN(glat) * sinalt;
  azsun:= FNatn2(y, x);

  if latitude<0 then altsun:=-altsun ;
  if abs(longitude)>90 then
    azsun:=azsun-180;

end;

{ ============================================================================= }
{ FUNCTION SunpositionAsVector                                                  }
{ Purpose : Given the upangle (altsun) and azimuth (angle from north) and the   }
{ distance of the sun return a vector with the suns position                    }
{ ============================================================================= }
function sunpositionAsVector(distance,azsun,altsun:single):TVector;
begin
  result[0]:=distance*sin(degtorad(azsun));
  result[1]:=distance*cos(degtorad(azsun));
  result[2]:=distance*sin(degtorad(altsun));
end;



end.
